@@ -1,5 +1,5 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from shlex import shlex
 from types import GeneratorType
 import gettext
@@ -7,12 +7,13 @@ import locale
 import decimal
 from decimal import Decimal
 import datetime
-import time
 import io
 from collections import OrderedDict
 
-from tryton.translate import date_format
-from tryton.common import untimezoned_date, datetime_strftime
+from tryton.common import untimezoned_date, timezoned_date, datetime_strftime
+from tryton.common.datetime_ import date_parse
+from tryton.common.timedelta import parse as timedelta_parse
+from tryton.common.timedelta import format as timedelta_format
 from tryton.pyson import PYSONDecoder
 
 __all__ = ['DomainParser']
@@ -117,6 +118,8 @@ def quote(value):
     "Quote string if needed"
     if not isinstance(value, basestring):
         return value
+    if '\\' in value:
+        value = value.replace('\\', '\\\\')
     if '"' in value:
         value = value.replace('"', '\\"')
     for test in (':', ' ', '(', ')') + OPERATORS:
@@ -129,6 +132,7 @@ def test_quote():
     assert quote('test') == 'test'
     assert quote('foo bar') == '"foo bar"'
     assert quote('"foo"') == '\\\"foo\\\"'
+    assert quote('foo\\bar') == 'foo\\\\bar'
 
 
 def ending_clause(domain, deep=0):
@@ -220,13 +224,16 @@ def test_split_target_value():
         assert split_target_value(field, value) == result
 
 
-def convert_value(field, value):
+def convert_value(field, value, context=None):
     "Convert value for field"
+    if context is None:
+        context = {}
 
     def convert_boolean():
         if isinstance(value, basestring):
             return any(test.decode('utf-8').lower().startswith(value.lower())
-                for test in (_('y'), _('yes'), _('true'), _('t'), '1'))
+                for test in (
+                    _('y'), _('Yes'), _('True'), _('t'), '1'))
         else:
             return bool(value)
 
@@ -256,31 +263,35 @@ def convert_value(field, value):
         return value
 
     def convert_datetime():
+        if not value:
+            return
+        format_ = context.get('date_format', '%x') + ' %X'
         try:
-            return untimezoned_date(datetime.datetime(*time.strptime(value,
-                        date_format() + ' ' + time_format(field))[:6]))
+            dt = date_parse(value, format_)
+            return untimezoned_date(dt)
         except ValueError:
-            try:
-                return datetime.datetime(*time.strptime(value,
-                        date_format())[:6])
-            except ValueError:
-                return
-        except TypeError:
             return
 
     def convert_date():
+        if not value:
+            return
+        format_ = context.get('date_format', '%x')
         try:
-            return datetime.date(*time.strptime(value, date_format())[:3])
+            return date_parse(value, format_).date()
         except (ValueError, TypeError):
             return
 
     def convert_time():
+        if not value:
+            return
         try:
-            return datetime.time(*time.strptime(value,
-                    time_format(field))[3:6])
+            return date_parse(value).time()
         except (ValueError, TypeError):
             return
 
+    def convert_timedelta():
+        return timedelta_parse(value, context.get(field.get('converter')))
+
     def convert_many2one():
         if value == '':
             return None
@@ -296,6 +307,7 @@ def convert_value(field, value):
         'datetime': convert_datetime,
         'date': convert_date,
         'time': convert_time,
+        'timedelta': convert_timedelta,
         'many2one': convert_many2one,
         }
     return converts.get(field['type'], lambda: value)()
@@ -389,13 +401,17 @@ def test_convert_datetime():
         'format': '"%H:%M:%S"',
         }
     for value, result in (
-            ('12/04/2002', datetime.datetime(2002, 12, 4)),
+            ('12/04/2002', untimezoned_date(datetime.datetime(2002, 12, 4))),
             ('12/04/2002 12:30:00', untimezoned_date(
                     datetime.datetime(2002, 12, 4, 12, 30))),
+            ('02/03/04', untimezoned_date(datetime.datetime(2004, 2, 3))),
+            ('02/03/04 05:06:07', untimezoned_date(
+                    datetime.datetime(2004, 2, 3, 5, 6, 7))),
             ('test', None),
             (None, None),
             ):
-        assert convert_value(field, value) == result
+        assert convert_value(field, value) == result, (value,
+            convert_value(field, value), result)
 
 
 def test_convert_date():
@@ -423,8 +439,22 @@ def test_convert_time():
         assert convert_value(field, value) == result
 
 
-def format_value(field, value, target=None):
+def test_convert_timedelta():
+    field = {
+        'type': 'timedelta',
+        }
+    for value, result in [
+            ('1d 2:00', datetime.timedelta(days=1, hours=2)),
+            ('foo', datetime.timedelta()),
+            (None, None),
+            ]:
+        assert convert_value(field, value) == result
+
+
+def format_value(field, value, target=None, context=None):
     "Format value for field"
+    if context is None:
+        context = {}
 
     def format_boolean():
         return _('True') if value else _('False')
@@ -458,22 +488,31 @@ def format_value(field, value, target=None):
     def format_datetime():
         if not value:
             return ''
-        format_ = date_format() + ' ' + time_format(field)
-        if (not isinstance(value, datetime.datetime)
-                or value.time() == datetime.time.min):
-            format_ = date_format()
-        return datetime_strftime(value, format_)
+        format_ = context.get('date_format', '%x') + ' ' + time_format(field)
+        if not isinstance(value, datetime.datetime):
+            time = datetime.datetime.combine(value, datetime.time.min)
+        else:
+            time = timezoned_date(value)
+        if time.time() == datetime.time.min:
+            format_ = '%x'
+        return datetime_strftime(time, format_)
 
     def format_date():
         if not value:
             return ''
-        return datetime_strftime(value, date_format())
+        format_ = context.get('date_format', '%x')
+        return datetime_strftime(value, format_)
 
     def format_time():
         if not value:
             return ''
         return datetime.time.strftime(value, time_format(field))
 
+    def format_timedelta():
+        if not value:
+            return ''
+        return timedelta_format(value, context.get(field.get('converter')))
+
     def format_many2one():
         if value is None:
             return ''
@@ -489,10 +528,11 @@ def format_value(field, value, target=None):
         'datetime': format_datetime,
         'date': format_date,
         'time': format_time,
+        'timedelta': format_timedelta,
         'many2one': format_many2one,
         }
     if isinstance(value, (list, tuple)):
-        return ';'.join(format_value(field, x) for x in value)
+        return ';'.join(format_value(field, x, context=context) for x in value)
     return quote(converts.get(field['type'],
             lambda: value if value is not None else '')())
 
@@ -587,8 +627,9 @@ def test_format_datetime():
         }
     for value, result in (
             (datetime.date(2002, 12, 4), '12/04/2002'),
-            (datetime.datetime(2002, 12, 4), '12/04/2002'),
-            (datetime.datetime(2002, 12, 4, 12, 30), '"12/04/2002 12:30:00"'),
+            (untimezoned_date(datetime.datetime(2002, 12, 4)), '12/04/2002'),
+            (untimezoned_date(datetime.datetime(2002, 12, 4, 12, 30)),
+                '"12/04/2002 12:30:00"'),
             (False, ''),
             (None, ''),
             ):
@@ -620,6 +661,19 @@ def test_format_time():
         assert format_value(field, value) == result
 
 
+def test_format_timedelta():
+    field = {
+        'type': 'timedelta',
+        }
+    for value, result in [
+            (datetime.timedelta(days=1, hours=2), '"1d 02:00"'),
+            (datetime.timedelta(), ''),
+            (None, ''),
+            ('', ''),
+            ]:
+        assert format_value(field, value) == result
+
+
 def complete_value(field, value):
     "Complete value for field"
 
@@ -632,7 +686,7 @@ def complete_value(field, value):
     def complete_selection():
         test_value = value if value is not None else ''
         if isinstance(value, list):
-            test_value = value[-1]
+            test_value = value[-1] or ''
         test_value = test_value.strip('%')
         for svalue, test in field['selection']:
             if test.lower().startswith(test_value.lower()):
@@ -655,7 +709,7 @@ def complete_value(field, value):
 
     def complete_datetime():
         yield datetime.date.today()
-        yield datetime.datetime.now()
+        yield datetime.datetime.utcnow()
 
     def complete_date():
         yield datetime.date.today()
@@ -688,6 +742,7 @@ def test_complete_selection():
             ('', ['male', 'female']),
             (None, ['male', 'female']),
             (['male', 'f'], [['male', 'female']]),
+            (['male', None], [['male', 'male'], ['male', 'female']]),
             ):
         assert list(complete_value(field, value)) == result
 
@@ -704,6 +759,25 @@ def test_complete_selection():
         assert list(complete_value(field_with_empty, value)) == result
 
 
+def test_complete_reference():
+    field = {
+        'type': 'reference',
+        'selection': [
+            ('spam', 'Spam'),
+            ('ham', 'Ham'),
+            ('', ''),
+            ],
+        }
+    for value, result in (
+            ('s', ['%spam%']),
+            ('test', []),
+            ('', ['%spam%', '%ham%', '%']),
+            (None, ['%spam%', '%ham%', '%']),
+            (['spam', 'h'], [['spam', 'ham']]),
+            ):
+        assert list(complete_value(field, value)) == result
+
+
 def parenthesize(tokens):
     "Nest tokens according to parenthesis"
     for token in tokens:
@@ -771,6 +845,8 @@ def test_operatorize():
     a = ('a', 'a', 'a')
     b = ('b', 'b', 'b')
     c = ('c', 'c', 'c')
+    null_ = ('d', None, 'x')
+    double_null_ = ('e', None, None)
     for value, result in (
             (['a'], ['a']),
             (['a', 'or', 'b'], [['OR', 'a', 'b']]),
@@ -796,6 +872,8 @@ def test_operatorize():
             (['a', iter(['b', 'or', 'c'])], ['a', [['OR', 'b', 'c']]]),
             ([a, iter([b, ('or',), c])], [a, [['OR', b, c]]]),
             (['a', iter(['b', 'or'])], ['a', [['OR', 'b']]]),
+            ([null_], [null_]),
+            ([null_, 'or', double_null_], [['OR', null_, double_null_]]),
             ):
         assert rlist(operatorize(iter(value))) == result
 
@@ -803,13 +881,14 @@ def test_operatorize():
 class DomainParser(object):
     "A parser for domain"
 
-    def __init__(self, fields):
+    def __init__(self, fields, context=None):
         self.fields = OrderedDict((name, f)
             for name, f in fields.iteritems()
             if f.get('searchable', True))
         self.strings = dict((f['string'].lower(), f)
             for f in fields.itervalues()
             if f.get('searchable', True))
+        self.context = context
 
     def parse(self, input_):
         "Return domain for the input string"
@@ -835,7 +914,24 @@ class DomainParser(object):
                         or isinstance(clause[0], (list, tuple)))
                     and all(isinstance(c, (list, tuple)) for c in clause[1:])):
                 return self.stringable(clause)
-            if clause[0] in self.fields or clause[0] == 'rec_name':
+            name, _, value = clause[:3]
+            if name.endswith('.rec_name'):
+                name = name[:-len('.rec_name')]
+            if name in self.fields:
+                field = self.fields[name]
+                if field['type'] in {
+                        'many2one', 'one2one', 'one2many', 'many2many'}:
+                    if field['type'] == 'many2one':
+                        types = (str, type(None))
+                    else:
+                        types = str
+                    if isinstance(value, (list, tuple)):
+                        return all(isinstance(v, types) for v in value)
+                    else:
+                        return isinstance(value, types)
+                else:
+                    return True
+            elif name == 'rec_name':
                 return True
             return False
 
@@ -880,10 +976,14 @@ class DomainParser(object):
                         operator = '!'
                     value = value.replace('%%', '%')
             def_operator = default_operator(field)
-            if (def_operator == operator.strip()
-                    or (def_operator in operator
-                        and ('not' in operator
-                            or '!' in operator))):
+            if def_operator == operator.strip():
+                operator = ''
+                if value in OPERATORS:
+                    # As the value could be interpreted as an operator,
+                    # the default operator must be forced
+                    operator = '"" '
+            elif (def_operator in operator
+                    and ('not' in operator or '!' in operator)):
                 operator = operator.rstrip(def_operator
                     ).replace('not', '!').strip()
             if operator.endswith('in'):
@@ -891,7 +991,7 @@ class DomainParser(object):
                     operator = '!'
                 else:
                     operator = ''
-            formatted_value = format_value(field, value, target)
+            formatted_value = format_value(field, value, target, self.context)
             if (operator in OPERATORS and
                     field['type'] in ('char', 'text', 'selection')
                     and value == ''):
@@ -965,6 +1065,8 @@ class DomainParser(object):
                 operator = None
             name = value
             value = ''
+        if not name:
+            name = ''
         if (name.lower() not in self.strings
                 and name not in self.fields):
             for field in self.strings.itervalues():
@@ -979,7 +1081,7 @@ class DomainParser(object):
             field = self.fields[name]
         else:
             field = self.strings[name.lower()]
-        if operator is None:
+        if not operator:
             operator = default_operator(field)
             value = ''
             if 'ilike' in operator:
@@ -1008,7 +1110,9 @@ class DomainParser(object):
                     else:
                         yield (None,)
                     name = (name,)
-                    if i + 1 < len(parts) and parts[i + 1] in OPERATORS:
+                    # empty string is also the default operator
+                    if (i + 1 < len(parts)
+                            and parts[i + 1] in OPERATORS + ('',)):
                         name += (parts[i + 1],)
                         i += 1
                     else:
@@ -1073,7 +1177,7 @@ class DomainParser(object):
                         if target:
                             field_name += '.rec_name'
 
-                    if operator is None:
+                    if not operator:
                         operator = default_operator(field)
                     if isinstance(value, list):
                         if operator == '!':
@@ -1086,20 +1190,21 @@ class DomainParser(object):
                             'datetime', 'date', 'time'):
                         if value and '..' in value:
                             lvalue, rvalue = value.split('..', 1)
-                            lvalue = convert_value(field, lvalue)
-                            rvalue = convert_value(field, rvalue)
+                            lvalue = convert_value(field, lvalue, self.context)
+                            rvalue = convert_value(field, rvalue, self.context)
                             yield iter([
                                     (field_name, '>=', lvalue),
-                                    (field_name, '<', rvalue),
+                                    (field_name, '<=', rvalue),
                                     ])
                             continue
                     if isinstance(value, list):
-                        value = [convert_value(field, v) for v in value]
+                        value = [convert_value(field, v, self.context)
+                            for v in value]
                         if field['type'] in ('many2one', 'one2many',
                                 'many2many', 'one2one'):
                             field_name += '.rec_name'
                     else:
-                        value = convert_value(field, value)
+                        value = convert_value(field, value, self.context)
                     if 'like' in operator:
                         value = likify(value)
                     if target:
@@ -1114,6 +1219,14 @@ def test_stringable():
                 'string': 'Name',
                 'type': 'char',
                 },
+            'relation': {
+                'string': 'Relation',
+                'type': 'many2one',
+                },
+            'relations': {
+                'string': 'Relations',
+                'type': 'many2many',
+                },
             })
     valid = ('name', '=', 'Doe')
     invalid = ('surname', '=', 'John')
@@ -1123,6 +1236,13 @@ def test_stringable():
     assert not dom.stringable(['AND', valid, invalid])
     assert dom.stringable([[valid]])
     assert not dom.stringable([[valid], [invalid]])
+    assert dom.stringable([('relation', '=', None)])
+    assert dom.stringable([('relation', '=', "Foo")])
+    assert dom.stringable([('relation.rec_name', '=', "Foo")])
+    assert not dom.stringable([('relation', '=', 1)])
+    assert dom.stringable([('relations', '=', "Foo")])
+    assert dom.stringable([('relations', 'in', ["Foo"])])
+    assert not dom.stringable([('relations', 'in', [42])])
 
 
 def test_string():
@@ -1167,6 +1287,7 @@ def test_string():
     assert dom.string([('name', '=', '')]) == 'Name: =""'
     assert dom.string([('name', 'ilike', '%')]) == 'Name: '
     assert dom.string([('name', 'ilike', '%Doe%')]) == 'Name: Doe'
+    assert dom.string([('name', 'ilike', '%<%')]) == 'Name: "" "<"'
     assert dom.string([('name', 'ilike', 'Doe')]) == 'Name: =Doe'
     assert dom.string([('name', 'ilike', 'Doe%')]) == 'Name: Doe%'
     assert dom.string([('name', 'ilike', 'Doe%%')]) == 'Name: =Doe%'
@@ -1295,6 +1416,9 @@ def test_group():
     assert rlist(dom.group(udlex(u'Name: \\"foo\\"'))) == [
         ('Name', None, '"foo"'),
         ]
+    assert rlist(dom.group(udlex(u'Name: "" <'))) == [
+        ('Name', '', '<'),
+        ]
 
 
 def test_parse_clause():
@@ -1337,6 +1461,8 @@ def test_parse_clause():
         ('rec_name', 'ilike', '%John%')]
     assert rlist(dom.parse_clause([('Name', None, None)])) == [
         ('name', 'ilike', '%')]
+    assert rlist(dom.parse_clause([('Name', '', None)])) == [
+        ('name', 'ilike', '%')]
     assert rlist(dom.parse_clause([('Name', '=', None)])) == [
         ('name', '=', None)]
     assert rlist(dom.parse_clause([('Name', '=', '')])) == [
@@ -1366,7 +1492,7 @@ def test_parse_clause():
         ]
     assert rlist(dom.parse_clause([('Integer', None, '3..5')])) == [[
             ('integer', '>=', 3),
-            ('integer', '<', 5),
+            ('integer', '<=', 5),
             ]]
     assert rlist(dom.parse_clause([('Reference', None, 'foo')])) == [
         ('reference', 'ilike', '%foo%'),
@@ -1380,9 +1506,31 @@ def test_parse_clause():
     assert rlist(dom.parse_clause([('Reference', None, ['foo', 'bar'])])) == [
         ('reference', 'in', ['foo', 'bar']),
         ]
+    assert rlist(dom.parse_clause(['OR',
+                ('Name', None, 'John'), ('Name', None, 'Jane')])) == ['OR',
+        ('name', 'ilike', '%John%'),
+        ('name', 'ilike', '%Jane%'),
+        ]
     assert rlist(dom.parse_clause([('Many2One', None, 'John')])) == [
         ('many2one', 'ilike', '%John%'),
         ]
     assert rlist(dom.parse_clause([('Many2One', None, ['John', 'Jane'])])) == [
         ('many2one.rec_name', 'in', ['John', 'Jane']),
         ]
+
+
+def test_completion():
+    dom = DomainParser({
+            'name': {
+                'string': 'Name',
+                'name': 'name',
+                'type': 'char',
+                },
+            })
+    assert list(dom.completion(u'Nam')) == ['Name: ']
+    assert list(dom.completion(u'Name:')) == ['Name: ']
+    assert list(dom.completion(u'Name: foo')) == []
+    assert list(dom.completion(u'Name: !=')) == []
+    assert list(dom.completion(u'Name: !=foo')) == []
+    assert list(dom.completion(u'')) == ['Name: ']
+    assert list(dom.completion(u' ')) == ['', 'Name: ']
